
/* This is an independent implementation of the DEAL encryption		*/
/* algorithm submitted by Richard Outerbridge and Lars Knudsen as	*/
/* a candidate the NIST AES programme.								*/
/*																	*/
/* Copyright in this implementation is held by Dr B R Gladman but	*/
/* I hereby give permission for its free direct or derivative use	*/
/* subject to aks_keynowledgment of its origin and compliance with	*/
/* any constraints that are placed on the exploitation of DEAL by	*/
/* its designers.													*/
/*																	*/
/* Dr Brian Gladman (gladman@seven77.demon.co.uk) 30th Noveember 98	*/

/* Timing Data:

Algorithm: deal (deal0.c)

128 bit key:
Key Setup:    8777 cycles
Encrypt:      2833 cycles =     9.0 mbits/sec
Decrypt:      2836 cycles =     9.0 mbits/sec
Mean:         2835 cycles =     9.0 mbits/sec
192 bit key:
Key Setup:    8812 cycles
Encrypt:      2833 cycles =     9.0 mbits/sec
Decrypt:      2833 cycles =     9.0 mbits/sec
Mean:         2833 cycles =     9.0 mbits/sec
256 bit key:
Key Setup:   11745 cycles
Encrypt:      3758 cycles =     6.8 mbits/sec
Decrypt:      3757 cycles =     6.8 mbits/sec
Mean:         3757 cycles =     6.8 mbits/sec

*/


#include "des.h"
#include "crypto_api.h"

/* These are all based on 32 bit unsigned values and may therefore  */
/* require endian conversions                                       */

char **deal_cipher_name(void);
u4byte *deal_set_key(const u4byte in_key[], const u4byte key_len);
void deal_encrypt(const u4byte in_blk[4], u4byte out_blk[4]);
void deal_decrypt(const u4byte in_blk[4], u4byte out_blk[4]);

/* 3. Basic macros for speeding up generic operations               */

/* Circular rotate of 32 bit values                                 */

#ifndef _MSC_VER

#define rotr(x,n)   (((x) >> ((int)(n))) | ((x) << (32 - (int)(n))))
#define rotl(x,n)   (((x) << ((int)(n))) | ((x) >> (32 - (int)(n))))

#else

#include <stdlib.h>

#pragma intrinsic(_lrotr,_lrotl)
#define rotr(x,n)   _lrotr(x,n)
#define rotl(x,n)   _lrotl(x,n)

#endif

/* Invert byte order in a 32 bit variable                           */

#define bswap(x)    (rotl(x, 8) & 0x00ff00ff | rotr(x, 8) & 0xff00ff00)

/* Extract byte from a 32 bit quantity (little endian notation)     */ 

#ifndef byte
#define byte(x,n)   ((u1byte)((x) >> (8 * n)))
#endif
/* For inverting byte order in input/output 32 bit words if needed  */

#ifdef  BYTE_SWAP
#define io_swap(x)  bswap(x)
#else
#define io_swap(x)  (x)
#endif

/* For inverting the byte order of input/output blocks if needed    */

#ifdef  WORD_SWAP

#define get_block(x)                            \
    ((u4byte*)(x))[0] = io_swap(in_blk[3]);     \
    ((u4byte*)(x))[1] = io_swap(in_blk[2]);     \
    ((u4byte*)(x))[2] = io_swap(in_blk[1]);     \
    ((u4byte*)(x))[3] = io_swap(in_blk[0])

#define put_block(x)                            \
    out_blk[3] = io_swap(((u4byte*)(x))[0]);    \
    out_blk[2] = io_swap(((u4byte*)(x))[1]);    \
    out_blk[1] = io_swap(((u4byte*)(x))[2]);    \
    out_blk[0] = io_swap(((u4byte*)(x))[3])

#define get_key(x,len)                          \
    ((u4byte*)(x))[4] = ((u4byte*)(x))[5] =     \
    ((u4byte*)(x))[6] = ((u4byte*)(x))[7] = 0;  \
    switch((((len) + 63) / 64)) {               \
    case 2:                                     \
    ((u4byte*)(x))[0] = io_swap(in_key[3]);     \
    ((u4byte*)(x))[1] = io_swap(in_key[2]);     \
    ((u4byte*)(x))[2] = io_swap(in_key[1]);     \
    ((u4byte*)(x))[3] = io_swap(in_key[0]);     \
    break;                                      \
    case 3:                                     \
    ((u4byte*)(x))[0] = io_swap(in_key[5]);     \
    ((u4byte*)(x))[1] = io_swap(in_key[4]);     \
    ((u4byte*)(x))[2] = io_swap(in_key[3]);     \
    ((u4byte*)(x))[3] = io_swap(in_key[2]);     \
    ((u4byte*)(x))[4] = io_swap(in_key[1]);     \
    ((u4byte*)(x))[5] = io_swap(in_key[0]);     \
    break;                                      \
    case 4:                                     \
    ((u4byte*)(x))[0] = io_swap(in_key[7]);     \
    ((u4byte*)(x))[1] = io_swap(in_key[6]);     \
    ((u4byte*)(x))[2] = io_swap(in_key[5]);     \
    ((u4byte*)(x))[3] = io_swap(in_key[4]);     \
    ((u4byte*)(x))[4] = io_swap(in_key[3]);     \
    ((u4byte*)(x))[5] = io_swap(in_key[2]);     \
    ((u4byte*)(x))[6] = io_swap(in_key[1]);     \
    ((u4byte*)(x))[7] = io_swap(in_key[0]);     \
    }

#else

#define get_block(x)                            \
    ((u4byte*)(x))[0] = io_swap(in_blk[0]);     \
    ((u4byte*)(x))[1] = io_swap(in_blk[1]);     \
    ((u4byte*)(x))[2] = io_swap(in_blk[2]);     \
    ((u4byte*)(x))[3] = io_swap(in_blk[3])

#define put_block(x)                            \
    out_blk[0] = io_swap(((u4byte*)(x))[0]);    \
    out_blk[1] = io_swap(((u4byte*)(x))[1]);    \
    out_blk[2] = io_swap(((u4byte*)(x))[2]);    \
    out_blk[3] = io_swap(((u4byte*)(x))[3])

#define get_key(x,len)                          \
    ((u4byte*)(x))[4] = ((u4byte*)(x))[5] =     \
    ((u4byte*)(x))[6] = ((u4byte*)(x))[7] = 0;  \
    switch((((len) + 63) / 64)) {               \
    case 4:                                     \
    ((u4byte*)(x))[6] = io_swap(in_key[6]);     \
    ((u4byte*)(x))[7] = io_swap(in_key[7]);     \
    case 3:                                     \
    ((u4byte*)(x))[4] = io_swap(in_key[4]);     \
    ((u4byte*)(x))[5] = io_swap(in_key[5]);     \
    case 2:                                     \
    ((u4byte*)(x))[0] = io_swap(in_key[0]);     \
    ((u4byte*)(x))[1] = io_swap(in_key[1]);     \
    ((u4byte*)(x))[2] = io_swap(in_key[2]);     \
    ((u4byte*)(x))[3] = io_swap(in_key[3]);     \
    }

#endif

#ifdef  BLOCK_SWAP
#define BYTE_SWAP
#define WORD_SWAP
#endif

static char	*alg_name[] = { "deal", "deal0.c" };

char **deal_cipher_name()
{
	return alg_name;
}

u4byte  ks_key[2] = { 0x67452301, 0xefcdab89 };

u4byte	ks_ks[32];	/* storage for the local key schedule	*/
u4byte	ks_init = 0;

u4byte	k_len;		/* copy of key_length					*/ 
u4byte	l_key[256];	/* storage for key schedule				*/

/* initialise the key schedule from the user supplied key	*/

u4byte *deal_set_key(const u4byte key_blk[], const u4byte key_len)
{	u4byte	lk[2];

	if(!ks_init)
	{
		des_ky(ks_key, ks_ks); ks_init = 1;
	}

	k_len = (key_len + 63) / 64;

	switch(k_len)
	{
		case 2:
			lk[0] = key_blk[0]; 
			lk[1] = key_blk[1]; 
			des_ec(lk, lk, ks_ks); 
			des_ky(lk, l_key);

			lk[0] ^= key_blk[2]; 
			lk[1] ^= key_blk[3];
			des_ec(lk, lk, ks_ks); 
			des_ky(lk, l_key + 32);
			  
			lk[0] ^= key_blk[0] ^ 0x00000080;  
			lk[1] ^= key_blk[1];
			des_ec(lk, lk, ks_ks); 
			des_ky(lk, l_key + 64);
			  
			lk[0] ^= key_blk[2] ^ 0x00000040;  
			lk[1] ^= key_blk[3];
			des_ec(lk, lk, ks_ks); 
			des_ky(lk, l_key + 96);
			  
			lk[0] ^= key_blk[0] ^ 0x00000020;  
			lk[1] ^= key_blk[1];
			des_ec(lk, lk, ks_ks); 
			des_ky(lk, l_key + 128);
			  
			lk[0] ^= key_blk[2] ^ 0x00000010;  
			lk[1] ^= key_blk[3];
			des_ec(lk, lk, ks_ks); 
			des_ky(lk, l_key + 160);
			break;
		case 3:
			lk[0] = key_blk[0]; 
			lk[1] = key_blk[1]; 
			des_ec(lk, lk, ks_ks); 
			des_ky(lk, l_key);

			lk[0] ^= key_blk[2]; 
			lk[1] ^= key_blk[3];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 32);
			  
			lk[0] ^= key_blk[4];  
			lk[1] ^= key_blk[5];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 64);
			  
			lk[0] ^= key_blk[0] ^ 0x00000080;  
			lk[1] ^= key_blk[1];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 96);
			  
			lk[0] ^= key_blk[2] ^ 0x00000040;  
			lk[1] ^= key_blk[3];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 128);
			  
			lk[0] ^= key_blk[4] ^ 0x00000020;  
			lk[1] ^= key_blk[5];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 160);
			break;
		case 4:
			lk[0] = key_blk[0]; 
			lk[1] = key_blk[1]; 
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key);

			lk[0] ^= key_blk[2]; 
			lk[1] ^= key_blk[3];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 32);
			  
			lk[0] ^= key_blk[4];  
			lk[1] ^= key_blk[5];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 64);
			  
			lk[0] ^= key_blk[6];  
			lk[1] ^= key_blk[7];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 96);
			  
			lk[0] ^= key_blk[0] ^ 0x00000080;  
			lk[1] ^= key_blk[1];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 128);
			  
			lk[0] ^= key_blk[2] ^ 0x00000040;  
			lk[1] ^= key_blk[3];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 160);
			lk[0] ^= key_blk[4] ^ 0x00000020;  
			lk[1] ^= key_blk[5];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 192);
			  
			lk[0] ^= key_blk[6] ^ 0x00000010;  
			lk[1] ^= key_blk[7];
			des_ec(lk, lk, ks_ks); des_ky(lk, l_key + 224);
			break;
	}	
	return l_key;
}

/* encrypt a bloks_key of text	*/

void deal_encrypt(const u4byte in_blk[4], u4byte out_blk[4])
{	u4byte	a[2], b[2], c[2];

	a[0] = in_blk[0]; a[1] = in_blk[1];
	b[0] = in_blk[2]; b[1] = in_blk[3];

	des_ec(a, c, l_key +   0); b[0] ^= c[0]; b[1] ^= c[1];
	des_ec(b, c, l_key +  32); a[0] ^= c[0]; a[1] ^= c[1];

	des_ec(a, c, l_key +  64); b[0] ^= c[0]; b[1] ^= c[1];
	des_ec(b, c, l_key +  96); a[0] ^= c[0]; a[1] ^= c[1];

	des_ec(a, c, l_key + 128); b[0] ^= c[0]; b[1] ^= c[1];
	des_ec(b, c, l_key + 160); a[0] ^= c[0]; a[1] ^= c[1];

	if(k_len == 4)
	{
		des_ec(a, c, l_key + 192); b[0] ^= c[0]; b[1] ^= c[1];
		des_ec(b, c, l_key + 224); a[0] ^= c[0]; a[1] ^= c[1];
	}

	out_blk[0] = a[0]; out_blk[1] = a[1];
	out_blk[2] = b[0]; out_blk[3] = b[1];
}

/* decrypt a bloks_key of text	*/

void deal_decrypt(const u4byte in_blk[4], u4byte out_blk[4])
{	u4byte	a[2], b[2], c[2];

	a[0] = in_blk[0]; a[1] = in_blk[1];
	b[0] = in_blk[2]; b[1] = in_blk[3];

	if(k_len == 4)
	{
		des_ec(b, c, l_key + 224); a[0] ^= c[0]; a[1] ^= c[1];
		des_ec(a, c, l_key + 192); b[0] ^= c[0]; b[1] ^= c[1];
	}

	des_ec(b, c, l_key + 160); a[0] ^= c[0]; a[1] ^= c[1];
	des_ec(a, c, l_key + 128); b[0] ^= c[0]; b[1] ^= c[1];

	des_ec(b, c, l_key +  96); a[0] ^= c[0]; a[1] ^= c[1];
	des_ec(a, c, l_key +  64); b[0] ^= c[0]; b[1] ^= c[1];

	des_ec(b, c, l_key +  32); a[0] ^= c[0]; a[1] ^= c[1];
	des_ec(a, c, l_key +   0); b[0] ^= c[0]; b[1] ^= c[1];

	out_blk[0] = a[0]; out_blk[1] = a[1];
	out_blk[2] = b[0]; out_blk[3] = b[1];
}
